import sys
import types
import logging
import functools
import asyncio

from .callback import Callback, defer

logging.basicConfig(stream=sys.stderr,
                    format="%(message)s")
log = logging.getLogger("monocle")


class Return(object):
    def __init__(self, *args):
        # mimic the semantics of the return statement
        if len(args) == 0:
            self.value = None
        elif len(args) == 1:
            self.value = args[0]
        else:
            self.value = args

    def __repr__(self):
        return "<%s.%s object at 0x%x; value: %s>" % (self.__class__.__module__,
                                                      self.__class__.__name__,
                                                      id(self),
                                                      repr(self.value))


class InvalidYieldException(TypeError):
    pass


# @_o
def _o(f):
    @functools.wraps(f)
    def coroutine_wrapper(*a, **k):
        gen_f = f(*a, **k)
        if not isinstance(gen_f, types.GeneratorType):
            if isinstance(gen_f, Callback):
                return gen_f
            return defer(gen_f)

        # make a coroutine that runs the monocle oroutine generator,
        # translating yielded Callbacks and Returns to the way these
        # things are handled in asyncio
        @functools.wraps(f)
        async def gen_wrapper():
            r = None
            e = None
            while True:
                try:
                    if e is not None:
                        r = gen_f.throw(e)
                        e = None
                    else:
                        r = gen_f.send(r)
                except StopIteration:
                    return None

                if isinstance(r, Return):
                    return r.value

                try:
                    if isinstance(r, Callback):
                        r = await r.future
                    elif (asyncio.iscoroutine(r) or
                          isinstance(r, asyncio.Future)):
                        r = await r
                    else:
                        raise InvalidYieldException(
                            "Unexpected value '%s' of type '%s' yielded from o-routine '%s'.  "
                            "O-routines can only yield Callback, coroutine, Future, and Return types." % (r, type(r), f))
                except Exception as e2:
                    e = e2

        fut = asyncio.ensure_future(gen_wrapper())
        cb = Callback()

        def _trigger_callback_with_future(fut):
            e = fut.exception()
            if e:
                cb(e)
            else:
                cb(fut.result())

        fut.add_done_callback(_trigger_callback_with_future)
        return cb

    return coroutine_wrapper
o = _o


def log_exception(msg="", elide_internals=None):
    log.exception(msg)


@_o
def launch(oroutine, *args, **kwargs):
    try:
        cb = oroutine(*args, **kwargs)
        if (not isinstance(cb, (Callback, asyncio.Future)) and
            not asyncio.iscoroutine(cb)):
            yield Return(cb)

        r = yield cb
        yield Return(r)
    except GeneratorExit:
        raise
    except Exception:
        log_exception()



import asyncio
import collections


class Callback(collections.abc.Awaitable):
    """
    A monocle Callback object, implemented using asyncio.Future.
    """

    def __init__(self):
        self.future = asyncio.Future()

    def add(self, handler):
        if not callable(handler):
            raise TypeError("'%s' object is not callable" % type(handler).__name__)
        def callback_handler(fut):
            e = fut.exception()
            if e:
                handler(e)
            else:
                handler(fut.result())
        self.future.add_done_callback(callback_handler)

    def __call__(self, result):
        assert not hasattr(self, 'result'), "Already called back"
        self.result = result
        if isinstance(result, BaseException):
            self.future.set_exception(result)
        else:
            self.future.set_result(result)

    def __await__(self):
        return self.future.__await__()


def defer(result):
    cb = Callback()
    cb(result)
    return cb



import asyncio
import threading

from monocle import launch


class EventLoop(object):
    def __init__(self):
        self._event_loop = asyncio.get_event_loop()
        self._thread_ident = threading.get_ident()

    def queue_task(self, delay, callable, *args, **kw):
        def task():
            return launch(callable, *args, **kw)

        if threading.get_ident() != self._thread_ident:
            self._event_loop.call_soon_threadsafe(task)
        else:
            return self._event_loop.call_later(delay, task)

    def run(self):
        try:
            self._event_loop.run_forever()
        finally:
            self._event_loop.run_until_complete(
                self._event_loop.shutdown_asyncgens())
            self._event_loop.close()

    def halt(self):
        self._event_loop.stop()

evlp = EventLoop()
queue_task = evlp.queue_task
run = evlp.run
halt = evlp.halt



from . import core
from .core import _o, o, launch, Return, InvalidYieldException, log_exception

VERSION = '1.0'

_stack_name = None


def init(stack_name):
    global _stack_name
    _stack_name = stack_name
